/*
 *	SBaGen - Sequenced Binaural Beat Generator
 *
 *	(c) 1999-2005 Jim Peters <jim@uazu.net>.  All Rights Reserved.
 *	For latest version see http://sbagen.sf.net/ or
 *	http://uazu.net/sbagen/.  Released under the GNU GPL version 2.
 *      +---------------------+
 *	|Use at your own risk.|
 *	+---------------------+
 *	Code remastered by unknown corp.,Inc., generally for internal purpose
 *	by Ivan Karamazov under GNU GPL V2 terms.
 *
 *	" This program is free software; you can redistribute it and/or modify
 *	  it under the terms of the GNU General Public License as published by
 *	  the Free Software Foundation, version 2.
 *
 *	  This program is distributed in the hope that it will be useful,
 *	  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	  GNU General Public License for more details. "
 *
 *	See the file COPYING for details of this license.
 *
 *	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
 *
 *	Some code fragments in the Win32 audio handling are based on
 *	code from PLIB (c) 2001 by Steve Baker, originally released
 *	under the LGPL (slDSP.cxx and sl.h).  For the original source,
 *	see the PLIB project: http://plib.sf.net
 *
 *	The code for the Mac audio output was based on code from the
 *	FINK project's patches to ESounD, by Shawn Hsiao and Masanori
 *	Sekino.  See: http://fink.sf.net
 */
//	MP3 decoding using libMAD

#include "libs/mad.h"

/* read mp3 from the mix tread
*/
static int mp3_read(int *dst, int dlen);

static struct mad_stream stream;
static struct mad_frame frame;
static struct mad_synth synth;
static char *mp3_buf;
static int mp3_len;
static int synth_extra;		// Extra samples carried over from previous frame
static int mp3_fatal;		// fatal error flag, stop decoding

/* mad init and start the inbuf thread for MP3 stream
*/
void mp3_init(void)
{
 mp3_fatal = 0;

 mp3_len = 32768;
 mp3_buf = Alloc(mp3_len);

 // Setup MAD decoder
 mad_stream_init(&stream);
 mad_frame_init(&frame);
 mad_synth_init(&synth);
 mad_stream_options(&stream, 0);

 // Force first data read
 stream.error = MAD_ERROR_BUFLEN;
 synth_extra = 0;

 // GO!!
 inbuf_start(mp3_read, 256*1024);	// 1024K buffer: 3s@44.1kHz
}

/* mp3-related cleanup
*/
void mp3_term(void)
{
 mad_synth_finish(&synth);
 mad_frame_finish(&frame);
 mad_stream_finish(&stream);
 MFree(mp3_buf);
}

// Limiting and truncation to 20 bits
#define ROUND(xx) ((((xx)<-MAD_F_ONE) ? -MAD_F_ONE : 	\
		    ((xx)>=MAD_F_ONE) ? MAD_F_ONE-1 : 	\
		    (xx)) >> (MAD_F_FRACBITS-15-4))

/* read mp3 from the mix tread
*/
static int mp3_read(int *dst, int dlen)
{
 int *dst0 = dst;
 int cnt, a, val;

 //debug("mp3_read %d", dlen);
 // we can't work after fatal error
 if(mp3_fatal)
  return 0;

 while(dlen > 0)
 {
  // First use up any samples from previous synth frame
  if(synth_extra > 0)
  {
   cnt = dlen / 2;
   if(cnt > synth_extra)
    cnt = synth_extra;

   for(a = 0; a < cnt; a++)
   {
    int ii = synth.pcm.length - synth_extra--;

    val = synth.pcm.samples[0][ii];
    *dst++ = ROUND(val);

    val = synth.pcm.samples[1][ii];
    *dst++= ROUND(val);

    dlen -= 2;
   }
   continue;
  }

  // Load up more MP3 data from disk if required
  if(stream.error == MAD_ERROR_BUFLEN)
  {
   char *rp = mp3_buf;

   // Preserve start of last frame
   if(stream.next_frame)
   {
    cnt = stream.bufend - stream.next_frame;
    memmove(mp3_buf, stream.next_frame, cnt);
    rp += cnt;
   }
	 
   // Fill buffer
   cnt = fread(rp, 1, mp3_buf + mp3_len - rp, mix_in);
   if(cnt == 0)
   {
    if(feof(mix_in))
     return dst - dst0;
    // error("Read error on MP3 input stream:\n  %s", strerror(errno));
    // in Karamazov's code we can exit (or call error()) from this thread
    mp3_fatal = 1;
#ifndef UNDER_CE
    warn("Read error on MP3 input stream:\n  %s", strerror(errno));
#else
    warn("Read error on MP3 input stream, code is %08X", errno);
#endif
    return 0;
   }
   rp += cnt;

   mad_stream_buffer(&stream, (unsigned char *)mp3_buf, rp - mp3_buf);
   stream.error = 0;
  }

  // Decode frame
  if(mad_frame_decode(&frame, &stream))
  {
   if(MAD_RECOVERABLE(stream.error))
   {
    warn("MAD recoverable error: %s", mad_stream_errorstr(&stream));
    continue;
   }
   if(stream.error == MAD_ERROR_BUFLEN) 
    continue;

   // error("Fatal error decoding MP3 stream: %s", mad_stream_errorstr(&stream));
   // in Karamazov's code we can exit (or call error()) from this thread
   mp3_fatal = 1;
#ifndef UNDER_CE
   warn("Read error on MP3 input stream:\n  %s", strerror(errno));
#else
   warn("Read error on MP3 input stream, code is %08X", errno);
#endif
   return 0;
  }

  // Pickup sampling rate if not picked up yet
  if(out_rate_def)
  {
   out_rate = frame.header.samplerate;
   out_rate_def = 0;
  }

  // Synthesize frame into samples
  mad_synth_frame(&synth, &frame);
  synth_extra = synth.pcm.length;
 }

 return dst - dst0;
}

/* The End...
*/


